// code by jph

static char ANIMATED_BUFFER[8192];

class aniMated {
public:
  aniMated(void);
  virtual ~aniMated(void);
  void relax(void);                             // dont overload
  virtual void control(void) {};
  virtual void motion(void);                    // rarely overloaded! TODO
  virtual void draw(void);
  void draw_std(void);
  void shadow(tensor<double> pnt);
  virtual void draw_dyn(void);
  void setEllipsoid(double mas,tensor<double> len);
  void setCuboid(double mas,tensor<double> len);
  void setCylinder(double mas,tensor<double> len);
  void setTorus(double mas,double sca,int rat);
  static int collide(aniMated* ani0,aniMated* ani1);

  virtual char* toString(void);

  double&         mass(void);
  tensor<double>  momentum_lin(void);           // linear momentum
  tensor<double>  momentum_ang(void);           // angular momentum
  double          energy_kin(void);
  
  tensor<double>  link_lhs(tensor<double> vec,tensor<double> lev);
  tensor<double>  link_rhs1(tensor<double> vec);
  tensor<double>  link_rhs(tensor<double> vec);
  tensor<double>  link_ect(tensor<double> vec);
//----------------------------------------------------------------------
  tensor<double>  pos; // 4x4 state tensor (orientation, position)
  tensor<double>  vel; // 4   velocity WORLD
  tensor<double>  ome; // 4   angular velocity OBJECT
  tensor<double>  acc; // 4   accel OBJECT coordinates
  tensor<double>  tor; // 4   torque tensor 
  
  tensor<double>  fva;
  tensor<double>  fot;
  tensor<double>  ine; // 4x4 inertia tensor
  
  tensor<double>  col; // 4   color
  time            rmv;
  unsigned int    status;
//-------------------------------------------------------------------------------------------------    
  static double   dt;
  rapid           rap;
  meshgl*         msh;
  meshgl*         shd;
};

aniMated::aniMated(void) {
  pos=tensor<double>::eye(4);
  vel=tensor<double>::vector(4);
  ome=tensor<double>::vector(4);
  acc=tensor<double>::vector(4);
  tor=tensor<double>::vector(4);
  
  fva=tensor<double>::vector(4);
  fot=tensor<double>::vector(4);
  ine=tensor<double>::diag(tensor<double>::vector(4).as( 1.0, 1.0, 1.0, 1.0));
  
  col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, 1.0);
  msh=0;
  shd=0;
}

aniMated::~aniMated(void) {} // no news

void aniMated::relax(void) {
  acc=0;
  tor=0;
}

void aniMated::motion(void) { // input: acc (world), tor (object)
  assert(0<mass());
  // friction --> acc
  acc-=glI(pos)*vel%fva;  // for avatar
  tor-=ome%fot;           // for avatar
  // acc --> vel
  vel+=pos*acc*dt;
  ome+=ine.solve( tor - ome.x(ine*ome) )*dt; // tor (object)
  // vel --> pos
  pos.data[12]+=vel.data[0]*dt;
  pos.data[13]+=vel.data[1]*dt;
  pos.data[14]+=vel.data[2]*dt;  
  pos=glOrthogonalize(pos*glExp(ome*dt)); // TODO every iteration is a little exaggerated
  //pos=pos*glExp(ome*dt); // TODO every iteration is a little exaggerated
}

void aniMated::draw_std(void) {
  assert(pos.data);
  assert(col.data);
  glPushMatrix();
    glMultMatrixd(pos.data);
    glColor4dv(col.data);
    draw();
  glPopMatrix();
}

void aniMated::shadow(tensor<double> pnt) {
  if (shd) {
    glPushMatrix();
      glMultMatrixd(pos.data);    
      shd->shadow(pos.solve(pnt));
    glPopMatrix();
  }
}

void aniMated::draw(void) {
  if (msh) {
    msh->draw();
    //msh->drawN3F(0.2,0);
  }
}

void aniMated::draw_dyn(void) { // right now: call from world coords
  double alp=0.7;
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);
    glPushMatrix();
      glColor4d( 0.8, 0.8, 0.8, alp);
      glMultMatrixd(glMatrixFromVector(momentum_lin(),pos[3]).data);
      glutSolidCylinder(0,momentum_lin().norm(),.025,10);
    glPopMatrix();
    glPushMatrix();
      glColor4d( 1.0, 1.0, 0.0, alp);
      glMultMatrixd(glMatrixFromVector(momentum_ang(),pos[3]).data);
      glutSolidCylinder(0,momentum_ang().norm(),.025,10);
    glPopMatrix();
    //glPushMatrix();
    //  glMultMatrixd(pos.data);
    //  glScaled(ine(0,0),ine(1,1),ine(2,2));
    //  glDepthFunc(GL_LEQUAL); // TODO
    //  glLineWidth(1);
    //  glutWireSphere(1.005,(ine(0,0)+ine(2,2))*15,(ine(0,0)+ine(2,2)+ine(1,1)*2)*5);
    //  glDepthFunc(GL_LESS);    
    //glPopMatrix();
  glPopAttrib();  
}

void aniMated::setEllipsoid(double mas,tensor<double> len) {
  tensor<double> ext=tensor<double>::vector(4).as(len(0),len(1),len(2), 0.0);
  if (mas<TENSOR_EPS)
    mas=4*MATH_1_3*MATH_PI*ext(0)*ext(1)*ext(2);
  ine.diag( mas*(ext(1)*ext(1)+ext(2)*ext(2))/5.0, // ok
            mas*(ext(0)*ext(0)+ext(2)*ext(2))/5.0, // ok
            mas*(ext(0)*ext(0)+ext(1)*ext(1))/5.0, // ok
            mas);
  //rap.set(aniData::meshgl_sphere_14336.map(tensor<double>::diag(ext)));
}

void aniMated::setCuboid(double mas,tensor<double> len) {
  tensor<double> ext=tensor<double>::vector(4).as(len(0),len(1),len(2), 0.0);
  if (mas<TENSOR_EPS)
    mas=8*abs(ext(0)*ext(1)*ext(2));
  ine.diag( mas*(ext(1)*ext(1)+ext(2)*ext(2))/3.0, // ok
            mas*(ext(0)*ext(0)+ext(2)*ext(2))/3.0, // ok
            mas*(ext(0)*ext(0)+ext(1)*ext(1))/3.0, // ok
            mas);
  //rap.set(aniData::meshgl_cube_13824.map(tensor<double>::diag(ext)));
}

void aniMated::setCylinder(double mas,tensor<double> len) {
  tensor<double> ext=tensor<double>::vector(4).as(len(0),len(1),len(2), 0.0);
  if (mas<TENSOR_EPS)
    mas=MATH_PI*ext(0)*ext(1)*2*ext(2);
  ine.diag( mas*(ext(1)*ext(1)/4.0+ext(2)*ext(2)/3.0), // ok
            mas*(ext(0)*ext(0)/4.0+ext(2)*ext(2)/3.0), // ok
            mas*(ext(0)*ext(0)+ext(1)*ext(1))/4.0,     // ok
            mas);
  //rap.set(aniData::meshgl_cylinder_10408.map(tensor<double>::diag(ext)));
}

void aniMated::setTorus(double mas,double sca,int rat) {
  assert(rat==4); // TODO more meshes are needed
  double R=sca;
  double r=sca/rat;
  tensor<double> ext=tensor<double>::vector(4).as(sca,sca,sca, 0.0);
  if (mas<TENSOR_EPS)
    mas=2*MATH_PI*MATH_PI*R*r*r;       // ok
  ine.diag( mas*(r*r*5   +R*R*4)/8.0,  // ok
            mas*(r*r*5   +R*R*4)/8.0,  // ok
            mas*(r*r*0.75+R*R  ),      // ok
            mas);
  //rap.set(aniData::meshgl_torus_4_11326.map(tensor<double>::diag(ext)));
}

double& aniMated::mass(void) {
  return ine.data[15];
}

tensor<double> aniMated::momentum_lin(void) {
  return vel*mass();
}

tensor<double> aniMated::momentum_ang(void) {
  return pos*ine*ome+pos[3].x(vel)*mass();
}

double aniMated::energy_kin(void) {
  return 0.5*((vel|vel*mass())+(ine*ome|ome));
}

tensor<double> aniMated::link_lhs(tensor<double> vec,tensor<double> lev) {
  return gl3m(pos*glD(vec)*ine.solve(glD(lev)*glI(pos)));
}

tensor<double> aniMated::link_rhs1(tensor<double> vec) {
  return gl3v( pos * ( glD(ome)*ome.x(vec) + glD(vec)*ine.solve(ome.x(ine*ome) - tor ) + acc ));
}

tensor<double> aniMated::link_rhs(tensor<double> vec) {
  return gl3v( pos * ( glD(ome)*ome.x(vec) + glD(vec)*ine.solve(ome.x(ine*ome) - tor ) ));
}

tensor<double> aniMated::link_ect(tensor<double> vec) {
  return gl3m(pos*glD(vec)*ine.solve(glI(pos)));
}

int aniMated::collide(aniMated* ani0,aniMated* ani1) { // todo use & // TODO move to aniMated!
  return rapid::collide(ani0->pos,ani0->rap,ani1->pos,ani1->rap,RAPID_ALL_CONTACTS);
}

char* aniMated::toString(void) {
  sprintf(ANIMATED_BUFFER,"aniMated  mass %10.2e  E_kin %10.2e",mass(),energy_kin());
  return ANIMATED_BUFFER;
}

double aniMated::dt=0;
