// code by jph

// rotationally symmetric around z:
//    glutSolidCone
//    glutSolidDisc
//    glutSolidDiscbw

// always use RGBA (color4) instead of (color3)

// credits: shadow volumes after the paper by Cass Everitt and Mark J. Kildard,
//          Practical and Robust Stenciled Shadow Volumes for Hardware-Accelerated Rendering

//#define MESH_BAR		//include barycentric coordinate computations
//#define MESH_GL			  //include openGL display
//#define MESH_RAPID    
//#define MESH_GST		//include openGL display and phantom
//#define MESH_IO			//include mesh text file input and output
//#define MESH_DEBUG	//define to receive certain warnings

#include <windows.h>

#ifdef GLEW_INCLUDE
#include "../_extern/glew.h"
#endif

#include "../_extern/glut.h"
#include "_console.hpp"
#include "meshgl.hpp"
#include "rapid.hpp"

#define               KEY_ESC                 27
#define               KEY_DL                  127
#define               KEY_SPECIAL             256
#define               KEY_AL                  KEY_SPECIAL+100
#define               KEY_AR                  KEY_SPECIAL+102
#define               KEY_AU                  KEY_SPECIAL+101
#define               KEY_AD                  KEY_SPECIAL+103
#define               KEY_PU                  KEY_SPECIAL+104
#define               KEY_PD                  KEY_SPECIAL+105
#define               KEY_P1                  KEY_SPECIAL+106
#define               KEY_EN                  KEY_SPECIAL+107
#define               KEY_IN                  KEY_SPECIAL+108

tensor<int>           opengl_size=tensor<int>(1,2).as(640,480);
tensor<int>           opengl_youtube=tensor<int>(1,2).as(450,360);
tensor<int>           opengl_position=tensor<int>(1,2).as(50,50);
tensor<double>        opengl_camera=tensor<double>::eye(4);
tensor<double>        opengl_project(2,4,4);
tensor<double>        opengl_frustum=tensor<double>(1,2).as( 0.1, 1e40); // for shadows, use 1e50=\infty as far plane
tensor<char>          opengl_title=tensor<char>::string("");

tensor<int>           keyboard(1,512);
tensor<int>           keyboard_down(1,512);
time                  opengl_time;
time                  opengl_lnch;
time                  opengl_rate;

tensor<int>           mouse(1,2);
tensor<double>        mouse_drag(1,4);
linked<tensor<char>>  opengl_bulletin;
int                   opengl_bulletin_show=1;
int                   opengl_polygonmode=1;
int                   opengl_frame=0;
int                   opengl_shot=0;
int                   opengl_shot_num=0;            // used to enumerate the files:
int                   opengl_shot_sub=0;            // opengl_num-sub.rgba
int                   opengl_recording=0;
int                   opengl_recording_numel=40;    // total shots taken
int                   opengl_recording_frame=20;    // take shot every x frames
int                   opengl_recording_count=0;

tensor<char>          opengl_font("get/tensor/white11x07.font"); // courier 7x11 bold

tensor<double>        opengl_color_black   =tensor<double>(1,4).as( 0.0, 0.0, 0.0, 1.0);
tensor<double>        opengl_color_gray10  =tensor<double>(1,4).as( 0.1, 0.1, 0.1, 1.0);
tensor<double>        opengl_color_gray20  =tensor<double>(1,4).as( 0.2, 0.2, 0.2, 1.0);
tensor<double>        opengl_color_gray30  =tensor<double>(1,4).as( 0.3, 0.3, 0.3, 1.0);
tensor<double>        opengl_color_gray40  =tensor<double>(1,4).as( 0.4, 0.4, 0.4, 1.0);
tensor<double>        opengl_color_gray50  =tensor<double>(1,4).as( 0.5, 0.5, 0.5, 1.0);
tensor<double>        opengl_color_gray60  =tensor<double>(1,4).as( 0.6, 0.6, 0.6, 1.0);
tensor<double>        opengl_color_gray70  =tensor<double>(1,4).as( 0.7, 0.7, 0.7, 1.0);
tensor<double>        opengl_color_gray80  =tensor<double>(1,4).as( 0.8, 0.8, 0.8, 1.0);
tensor<double>        opengl_color_gray90  =tensor<double>(1,4).as( 0.9, 0.9, 0.9, 1.0);
tensor<double>        opengl_color_white   =tensor<double>(1,4).as( 1.0, 1.0, 1.0, 1.0);
tensor<double>        opengl_color_red     =tensor<double>(1,4).as( 1.0, 0.0, 0.0, 1.0);
tensor<double>        opengl_color_green   =tensor<double>(1,4).as( 0.0, 1.0, 0.0, 1.0);
tensor<double>        opengl_color_blue    =tensor<double>(1,4).as( 0.0, 0.0, 1.0, 1.0);
tensor<double>        opengl_color_cyan    =tensor<double>(1,4).as( 0.0, 1.0, 1.0, 1.0);
tensor<double>        opengl_color_magenta =tensor<double>(1,4).as( 1.0, 0.0, 1.0, 1.0);
tensor<double>        opengl_color_yellow  =tensor<double>(1,4).as( 1.0, 1.0, 0.0, 1.0);

tensor<double>        opengl_ambience      =opengl_color_gray20; // alpha does not have effect?

tensor<double>        opengl_clearcolor         =opengl_color_gray20;
int                   opengl_clearcolor_white   =0;

int                   opengl_lighting           =1;

class opengl_light {
public:
  opengl_light(void) {
    pos=tensor<double>::eye(4);
    dif=opengl_color_black;
    spc=opengl_color_black;
    lin=0;
    qdr=0;
    cut=180;
    exp=0;
    shadow=1;
    hash.add(count=counter++,this);
  }
  
  ~opengl_light(void) {
    hash.remove(count);
  }
  
  void enable(uint index) {
    glEnable(index);
    glLightfv(index,GL_POSITION,pos[3].to<float>().data);
	  glLightfv(index,GL_DIFFUSE ,dif.to<float>().data);
    glLightfv(index,GL_SPECULAR,spc.to<float>().data);
    glLightf(index,GL_LINEAR_ATTENUATION,lin);
    glLightf(index,GL_QUADRATIC_ATTENUATION,qdr);
    glLighti(index,GL_SPOT_CUTOFF,cut);
    glLightfv(index,GL_SPOT_DIRECTION,pos[0].to<float>().data);
    glLightf(index,GL_SPOT_EXPONENT,exp);
  }
  
  static void disable(void) {
    for (int c0=0;c0<8;c0++)
      glDisable(GL_LIGHT0+c0);
  }
  
  tensor<double>  pos; // 4x4
  tensor<double>  dif; // 4   affine
  tensor<double>  spc; // 4   affine
  double          lin;
  double          qdr;
  int             cut; // allowed values are: 0, 1 ... 90, and 180
  //tensor<double>  dir;
  double          exp;
  int             shadow;
  static hash<uint,opengl_light*> hash;
private:
  uint            count;
  static uint     counter;
};

uint opengl_light::counter=0;
hash<uint,opengl_light*> opengl_light::hash(32);

hash<tensor<char>,uint>       opengl_texture_hash(64);
hash<tensor<char>,meshgl*>    opengl_meshgl_hash(64);

// these functions require implementation:
void scene(void);     // animate objects, set modelview matrix
void display(void);   // draw meshes
void shadow(tensor<double> pos);    // draw shadow volumes (function body can be void)
void finish(void);

void opengl_iteration(void) {
  if (opengl_clearcolor_white)
    glClearColor(1,1,1,1);
  else
    glClearColor(opengl_clearcolor(0),opengl_clearcolor(1),opengl_clearcolor(2),opengl_clearcolor(3));

  glDepthMask(GL_TRUE);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  opengl_rate=time::now()-opengl_time-opengl_lnch;
  opengl_time+=opengl_rate.s();

  scene();

  // -------------------------------------- TESTING -----------
  /*
  tensor<uint> sel(1,5);
  glSelectBuffer(sel.numel,sel.data);
  glRenderMode(GL_SELECT);
  glInitNames();
  glPushName(0);
  tensor<int> vie(1,4);
  glGetIntegerv(GL_VIEWPORT,vie.data);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
    glLoadIdentity();
    gluPickMatrix(mouse(0),mouse(1),2,2,vie.data);
    glMultMatrixd(opengl_project.data);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
      glLoadIdentity();
      glMultMatrixd(opengl_camera.data);  
      
      display();
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  opengl_bulletin.add(tensor<char>::string( mouse.toString() ));
  glMatrixMode(GL_MODELVIEW);
  uint a=glRenderMode(GL_RENDER);
  opengl_bulletin.add(tensor<char>::string( sel.toString() ));
  opengl_bulletin.add(tensor<char>::string( vie.toString() ));
  opengl_bulletin.add(tensor<int>::string( a ));
  */
  // -------------------------------------- TESTING -----------
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glMultMatrixd(opengl_camera.data);  
  glDepthFunc(GL_LESS);
  glCullFace(GL_BACK);
  if (opengl_lighting)
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT,opengl_ambience.to<float>().data);
  else
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT,opengl_color_white.to<float>().data);

  int cnt=0;
  int itr=0;
  tensor<opengl_light*> lgt=opengl_light::hash.toTensor();
  for (itr=0;itr<lgt.numel;++itr) {
    if (opengl_lighting&&!lgt(itr)->shadow)
      lgt(itr)->enable(GL_LIGHT0+cnt++);
    if (cnt==8)
      break;
  }
  
  display(); opengl_light::disable(); cnt=0;
  
  //for (int c0=0;c0<lgt.numel*opengl_lighting;++c0) {
  //  glPushMatrix();  //    glMultMatrixd(lgt(c0)->pos.data);  //    glColor4dv(lgt(c0)->dif.data);
  //    //glutWireSphere( 0.1,10,10);  //  glPopMatrix();  //}
  
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,opengl_color_black.to<float>().data);    
  glDepthMask(GL_FALSE);
  
  if (opengl_lighting) {
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE,GL_ONE);

    glDepthFunc(GL_EQUAL);    
    for (;itr<lgt.numel;++itr) {
      if (!lgt(itr)->shadow)
        lgt(itr)->enable(GL_LIGHT0+cnt++);
      if (cnt==8) {
        display(); opengl_light::disable(); cnt=0;
      }
    }
    if (cnt) {
      display(); opengl_light::disable(); cnt=0;
    }
    glDepthFunc(GL_LESS);
    
    glEnable(GL_STENCIL_TEST);
    for (int c0=0;c0<lgt.numel;++c0)
      if (lgt(c0)->shadow) {
        glClear(GL_STENCIL_BUFFER_BIT);
        glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE);
        glStencilFunc(GL_ALWAYS,0,~0);
        glStencilMask(~0);
        
        shadow(lgt(c0)->pos);
        
        lgt(c0)->enable(GL_LIGHT0);
        glStencilFunc(GL_EQUAL,0,~0);
        glStencilOp(GL_KEEP,GL_KEEP,GL_INCR);
        glDepthFunc(GL_EQUAL);
        glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);  
        display();
        glDepthFunc(GL_LESS);     
        glDisable(GL_LIGHT0);
      }       
    glDisable(GL_STENCIL_TEST);    
    
    glDisable(GL_BLEND);
  }
  
  finish();

  opengl_bulletin.add(tensor<char>::string(opengl_light::hash.toInfo()));
  opengl_bulletin.add(tensor<char>::string("rate ")&tensor<double>::string( 1/opengl_rate.s() )&tensor<char>::string("time ")&tensor<double>::string( opengl_time.s() ));

  if (opengl_bulletin_show) {
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
      glLoadIdentity();
      gluOrtho2D(0,opengl_size(0),0,opengl_size(1));    
      glPushAttrib(GL_ENABLE_BIT);
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);           
        int piy=opengl_size(1);
        int inc=opengl_font.cumsize().last();
        tensor<char> str;
        while (opengl_bulletin.token(str)) {
          int pix=0;
          piy-=opengl_font.size[1];
          for (int c0=0;c0<str.numel;c0++) {
            glRasterPos2i(pix,piy);
            switch ((int)str(c0)) {
              case 10:
                pix=0;
                piy-=opengl_font.size[1];
                break;
              default:
                pix+=opengl_font.size[2];
                int c1=str(c0);
                if (32<c1&&c1<128&&!keyboard('u'))
                  glDrawPixels(opengl_font.size[2],opengl_font.size[1],GL_RGBA,GL_UNSIGNED_BYTE,opengl_font^(c1-32)*inc);
            }
          }
          opengl_bulletin.remove();
        }
      glPopAttrib();
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
  }
  opengl_bulletin.empty();

  if (keyboard('g')) {          //opengl control mode
    char opengl_txt[200];
    sprintf(opengl_txt,"opengl\n (b)ulletin =%i\n (c)learcolor white =%i\n (l)ighting effects =%i\n (s)creen shot\n (r)ecording\n (w)ireframe =%i",opengl_bulletin_show,opengl_clearcolor_white,opengl_lighting,opengl_polygonmode);
    opengl_bulletin.add(tensor<char>::string( opengl_txt ));
    opengl_bulletin_show^=keyboard_down('b');
    opengl_lighting^=keyboard_down('l');
    opengl_clearcolor_white^=keyboard_down('c');
    opengl_shot|=keyboard_down('s');
    opengl_recording|=keyboard_down('r');
    if (keyboard_down('w')) {   //wireframe
      opengl_polygonmode^=1;
      if (opengl_polygonmode)
        glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      else
        glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);      
    }
  }
  
  if (opengl_shot_sub>=opengl_recording_numel) {
    opengl_recording=false;
    opengl_shot_num++;
    opengl_shot_sub=0;
  }
  if (opengl_recording)
    opengl_shot=!(opengl_recording_count++%opengl_recording_frame);
  else
    opengl_recording_count=0;
  
  if (opengl_shot) {
    tensor<uchar> rgba(3,4,opengl_size(0),opengl_size(1));
    tensor<float> zbuf(2,opengl_size(0),opengl_size(1));
    tensor<uchar> stnc(2,opengl_size(0),opengl_size(1));
    glReadPixels(0,0,opengl_size(0),opengl_size(1),GL_RGBA,GL_UNSIGNED_BYTE   ,rgba.data);
    //glReadPixels(0,0,opengl_size(0),opengl_size(1),GL_DEPTH_COMPONENT,GL_FLOAT,zbuf.data);
    //glReadPixels(0,0,opengl_size(0),opengl_size(1),GL_STENCIL_INDEX,GL_UNSIGNED_BYTE,stnc.data);
    char fnc[80];
    sprintf(fnc,"out/opengl_%02i-%04i_rgba.bmp",opengl_shot_num,opengl_shot_sub);
    format::bmp(fnc,rgba);    
    //sprintf(fnc,"out/opengl_%02i-%04i_zbuf.mat",opengl_shot_num,opengl_shot_sub);
    //zbuf.store(fnc);    
    //sprintf(fnc,"out/opengl_%02i-%04i_stnc.mat",opengl_shot_num,opengl_shot_sub);
    //stnc.store(fnc);
    
    ++opengl_shot_sub;
    opengl_shot=0;
  }
  keyboard_down=0;
  mouse_drag=0;
  ++opengl_frame;
  
  glutSwapBuffers();
}

void opengl_idle(void) {
  glutPostRedisplay();
}

void opengl_keyboard(uchar key,int pix,int piy) {  
  //int mod=glutGetModifiers();
  keyboard(key)=1;
  keyboard_down(key)=1;
}

void opengl_keyboardUp(uchar key,int pix,int piy) {  
  if (key==KEY_ESC)
    exit(0);
  keyboard(key)=0;
}

void opengl_special(int key,int pix,int piy) {  
  keyboard(KEY_SPECIAL+key)=1;
  keyboard_down(KEY_SPECIAL+key)=1;
}

void opengl_specialUp(int key,int x,int y) {
  keyboard(KEY_SPECIAL+key)=0;
}

void opengl_mouse(int b,int s,int x,int y) {
  mouse.as(x,y);
}

void opengl_motion(int x,int y) {
  mouse_drag.as(
    (x-mouse(0))/(double)opengl_size(0),
    (mouse(1)-y)/(double)opengl_size(1),
     0.0,
     0.0);
  mouse.as(x,y);
}

void opengl_passiveMotion(int x,int y) {
  mouse.as(x,y);
}

void opengl_entry(int entry) {
#ifndef NDEBUG
  if (!entry)
    exit(0);
#endif
}

void opengl_reshape(int w,int h) { // ok (also called at launch)
  opengl_size.as(w,h);
  glViewport(0,0,w,h);
//  tensor<int> vie(1,4);
//  glGetIntegerv(GL_VIEWPORT,vie.data);  
  glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //gluPickMatrix(mouse(0),mouse(1),100,100,vie.data);
    gluPerspective(55,w/(float)h,opengl_frustum(0),opengl_frustum(1));
    glGetDoublev(GL_PROJECTION_MATRIX,opengl_project.data);
  opengl_iteration();
}

void opengl_main(int argc,char** argv) {
  glutInit(&argc,argv);
  glutInitDisplayMode(GLUT_DOUBLE|GLUT_DEPTH|GLUT_RGBA|GLUT_STENCIL);
#ifdef OPENGL_FULLSCREEN
  glutGameModeString(OPENGL_FULLSCREEN);
  if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE))
    glutEnterGameMode();
  else {
    printf("OPENGL_FULLSCREEN error!\n");
    exit(0);
  }
#else
  glutInitWindowSize(opengl_size(0),opengl_size(1));
  glutInitWindowPosition(opengl_position(0),opengl_position(1));
  glutCreateWindow((opengl_title&0).data);
#endif

#ifdef GLEW_INCLUDE
  if (GLEW_OK!=glewInit())
    printf("Warning: GLEW is not supported...\n");
  printf("Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));  
  if (GLEW_ARB_vertex_shader && GLEW_ARB_fragment_shader)
	  printf(" -> ready for GLSL\n");
#endif
  
  glutDisplayFunc(opengl_iteration);                //callbacks
  glutIdleFunc(opengl_idle);
  glutKeyboardFunc(opengl_keyboard);
  glutKeyboardUpFunc(opengl_keyboardUp);
  glutSpecialFunc(opengl_special);
  glutSpecialUpFunc(opengl_specialUp);
  glutMouseFunc(opengl_mouse);
  glutMotionFunc(opengl_motion);
  glutPassiveMotionFunc(opengl_passiveMotion);
  glutEntryFunc(opengl_entry);
  glutReshapeFunc(opengl_reshape);
  glutIgnoreKeyRepeat(true);
  
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glColorMaterial(GL_FRONT,GL_SPECULAR);
  glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,1); // specular light: with respect to 0-> z axis, 1 -> observer
  
  //glShadeModel(GL_SMOOTH);              //GL_FLAT
  //glPixelStorei(GL_UNPACK_ALIGNMENT,1); //texture
  //glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
  //glClearDepth(1); // specify the clear value for the depth buffer
  
  //float fog_color[4]={ 0.0, 0.0, 1.0, 1.0};
  //glFogi(GL_FOG_MODE, GL_EXP); // GL_LINEAR, GL_EXP2 Fog Mode
  //glFogfv(GL_FOG_COLOR, fog_color); // Set Fog Color
  //glFogf(GL_FOG_DENSITY, 0.05f); // How Dense Will The Fog Be
  //glHint(GL_FOG_HINT, GL_NICEST); // Fog Hint Value
  //glFogf(GL_FOG_START, 1.0f); // Fog Start Depth
  //glFogf(GL_FOG_END, 100.0f); // Fog End Depth
  //glEnable(GL_FOG); // Enables GL_FOG  
  
  //glEnable(GL_LINE_SMOOTH); does not look good if drawing order is not ordered
  //glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
  //glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  
  opengl_lnch=time::now();
#ifdef NDEBUG
  opengl_bulletin_show=0;
#endif
}

void opengl_grid() {
  glNormal3d( 0.0, 0.0, 1.0);
  double u= 0.0; // shift down along z a tiny bit
  glLineWidth(2);
  glColor4d( 0.7, 0.4, 0.0, 0.8);
  glBegin(GL_LINES);
    glVertex3d(-20.0,  0.0,  u  ); glVertex3d( 20.0,  0.0,  u  );
    glVertex3d(  0.0,-20.0,  u  ); glVertex3d(  0.0, 20.0,  u  );
    glVertex3d(  0.0,  0.0,-20.0); glVertex3d(  0.0,  0.0, 20.0);
  glEnd();
  glColor4d( 0.3, 0.7, 0.3, 0.8);
  glLineWidth(1);
  glBegin(GL_LINES);
    for (int c0=-10;c0<11;++c0) {
      glVertex3d(-10.0, c0  ,  u); glVertex3d( 10.0, c0  ,  u);
      glVertex3d( c0  ,-10.0,  u); glVertex3d( c0  , 10.0,  u);
    }
  glEnd();
  if (0) {
    u= 50.0;
    double h=-50.0;
    glColor4d( 1.0, 1.0, 1.0, 1.0);
    glBegin(GL_QUADS);
      glVertex3d(-u, u, h);
      glVertex3d(-u,-u, h);
      glVertex3d( u,-u, h);
      glVertex3d( u, u, h);
    glEnd();
  }
}

void opengl_grid_xy() {
  assert(false);
  glPushAttrib(GL_LIGHTING_BIT);
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    //glHint(GL_LINE_SMOOTH_HINT,GL_DONT_CARE); // GL_DONT_CARE
    glColor4d( 0.3, 0.3, 0.3, 1.0);
    glLineWidth(1);
    glBegin(GL_LINES);
      for (int c0=-10;c0<=10;++c0) {
        glVertex3f(-10, c0, -5);
        glVertex3f( 10, c0, -5);
        glVertex3f( c0,-10, -5);
        glVertex3f( c0, 10, -5);
      }
    glEnd();
    //glHint(GL_LINE_SMOOTH_HINT,GL_NICEST); // GL_DONT_CARE
  glPopAttrib();
  glPopAttrib();
}

uint opengl_texture(char* str) {
  uint ret;
  tensor<char> key=tensor<char>::string(str);
  if (!opengl_texture_hash.get(key,ret)) {
    glGenTextures(1,&ret);
    assert(ret); // 0 is not allowed in our world
    glBindTexture(GL_TEXTURE_2D,ret);
    glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT); // this is default
    glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT); // this is default
    tensor<uchar> tex(str);
    assert(tensor<char>::valid);
    gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA,tex.size[1],tex.size[2],GL_RGBA,GL_UNSIGNED_BYTE,tex.data);      
    opengl_texture_hash.add(key,ret);    
    //char fnc[80];
    //sprintf(fnc,"out/texture_%i.bmp",ret);
    //format::bmp(fnc,tex);
  }
  return ret;
}

meshgl* opengl_meshgl(char* str) {
  meshgl* ret=0;
  tensor<char> key=tensor<char>::string(str);
  if (!opengl_meshgl_hash.get(key,ret)) {
    ret=new meshgl(str);
    opengl_meshgl_hash.add(key,ret);
  }
  return ret;
}

void opengl_meshgl_delete(void) {
  //opengl_meshgl_hash.empty_delete();
}

tensor<double> glD(const tensor<double>& myTensor) { // ok
  assert(myTensor.numel==3||myTensor.numel==4);
  return tensor<double>(2,4,4).as(
    0.,myTensor.data[2],-myTensor.data[1],0.,
    -myTensor.data[2],0.,myTensor.data[0],0.,
    myTensor.data[1],-myTensor.data[0],0.,0.,
    0.,0.,0.,0.);
}

tensor<double> glT(const tensor<double>& myTensor) { // ok
  assert(3<=myTensor.numel);
  return tensor<double>(2,4,4).as(
    1.,0.,0.,0.,
    0.,1.,0.,0.,
    0.,0.,1.,0.,
    myTensor.data[0],myTensor.data[1],myTensor.data[2],1.);
}

tensor<double> glI(const tensor<double>& myTensor) { // ok
  return tensor<double>(2,4,4).as(
     myTensor.data[ 0],myTensor.data[ 4],myTensor.data[ 8],0.,
     myTensor.data[ 1],myTensor.data[ 5],myTensor.data[ 9],0.,
     myTensor.data[ 2],myTensor.data[ 6],myTensor.data[10],0.,
    -myTensor.data[12]*myTensor.data[ 0]-myTensor.data[13]*myTensor.data[ 1]-myTensor.data[14]*myTensor.data[ 2],
    -myTensor.data[12]*myTensor.data[ 4]-myTensor.data[13]*myTensor.data[ 5]-myTensor.data[14]*myTensor.data[ 6],
    -myTensor.data[12]*myTensor.data[ 8]-myTensor.data[13]*myTensor.data[ 9]-myTensor.data[14]*myTensor.data[10],1.);
}

tensor<double> glN(tensor<double> myTensor) {
  tensor<double> nrm=myTensor.normal();
  return tensor<double>(2,4,4).as(
    nrm.data[ 0]*nrm.data[ 0],nrm.data[ 1]*nrm.data[ 0],nrm.data[ 2]*nrm.data[ 0],0.,
    nrm.data[ 0]*nrm.data[ 1],nrm.data[ 1]*nrm.data[ 1],nrm.data[ 2]*nrm.data[ 1],0.,
    nrm.data[ 0]*nrm.data[ 2],nrm.data[ 1]*nrm.data[ 2],nrm.data[ 2]*nrm.data[ 2],0.,
    0.,0.,0.,1.);
}

tensor<double> gl3v(tensor<double> myTensor) {
  return tensor<double>(1,3).as(
    myTensor.data[0],
    myTensor.data[1],
    myTensor.data[2]);
}

tensor<double> gl4v(tensor<double> myTensor) {
  return tensor<double>(1,4).as(
    myTensor.data[0],
    myTensor.data[1],
    myTensor.data[2],
    0.0);
}

tensor<double> gl3m(tensor<double> myTensor) { // ok
  return tensor<double>(2,3,3).as(
    myTensor.data[ 0],myTensor.data[ 1],myTensor.data[ 2], // appears in code as 'transposed'
    myTensor.data[ 4],myTensor.data[ 5],myTensor.data[ 6],    
    myTensor.data[ 8],myTensor.data[ 9],myTensor.data[10]);
}

tensor<double> gl4r(tensor<double> myTensor) { // ok
  return tensor<double>(2,4,4).as(
    myTensor.data[ 0],myTensor.data[ 1],myTensor.data[ 2],0.0, // appears in code as 'transposed'
    myTensor.data[ 4],myTensor.data[ 5],myTensor.data[ 6],0.0,    
    myTensor.data[ 8],myTensor.data[ 9],myTensor.data[10],0.0,
    0.0,0.0,0.0,1.0);
}

tensor<double> glOrthogonalize(tensor<double> pos) {
  tensor<double> v0=pos[0].normal();
  tensor<double> v1=pos[1]-v0*(v0|pos[1]);                v1=v1.normal();
  tensor<double> v2=pos[2]-v0*(v0|pos[2])-v1*(v1|pos[2]); v2=v2.normal();
  return tensor<double>(2,4,4).as(
    v0(0),v0(1),v0(2), 0.0,
    v1(0),v1(1),v1(2), 0.0,
    v2(0),v2(1),v2(2), 0.0,
    pos(0,3),pos(1,3),pos(2,3), 1.0);
}

tensor<double> glMatrixFromVector(tensor<double> dir,tensor<double> pos) {
  if (TENSOR_EPS<dir.norm()) {
    tensor<double> vc0=gl3v(dir);
    vc0=vc0.normal();
    int ind;
    vc0.absolute().min_(ind);
    tensor<double> vc1(1,3);
    vc1.data[ind]=1;
    vc1=vc1-vc0*(vc0|vc1);
    vc1=vc1.normal();
    tensor<double> ret(2,4,4);
    ret.in( 0,vc1);
    ret.in( 4,vc0.x(vc1));
    ret.in( 8,vc0);
    ret.in(12,pos.data,3);
    ret.data[15]=1;
    return ret;
  } else
    return tensor<double>::eye(4);
}

tensor<double> glExp(tensor<double> ome) {
  tensor<double> mat=glD(ome);
  double fac=ome|ome; // norm squared
  double sna,cna;
  sna=+1.0+fac*(
      -1/6.0+fac*(
      +1/120.0+fac*(
      -1/5040.0+fac*(
      +1/362880.0+fac*(
      -1/39916800.0+fac*(
      +1/6227020800.0+fac*(
      -1/1307674368000.0
      )))))));
  cna=+1/2.0+fac*(
      -1/24.0+fac*(
      +1/720.0+fac*(
      -1/40320.0+fac*(
      +1/3628800.0+fac*(
      -1/479001600.0+fac*(
      +1/87178291200.0+fac*(
      -1/20922789888000.0
      )))))));
  return tensor<double>::eye(4)+mat*sna+mat*mat*cna;
}

// glMatchVectors returns 4x4 matrix R so that  R.beg == end
// if beg == end then R is the identity
tensor<double> glMatchVectors(tensor<double> beg,tensor<double> end) {
  tensor<double> crs=beg.x(end);
  double fac=asin(crs.norm());
  fac*=fac;
  crs*=1.0+fac*(
       1/6.0+fac*(
       7/360.0+fac*(
       31/15120.0+fac*(
       127/604800.0+fac*(
       73/3421440.0+fac*(
       1414477/653837184000.0+fac*(
       8191/37362124800.0+fac*
       16931177/762187345920000.0
       )))))));
  return glExp(crs);
}



// =========================================================================
// DRAWING FUNCTIONS

// 2D -----------------------------------

void glutSolidDisc(double rad,int slices) { // ok
  double rab=abs(rad);
  double fac=MATH_2PI/slices;
  glBegin(GL_TRIANGLE_FAN);
    glNormal3d(0,0,math::Sign(rad));
    glVertex3d(0,0,0);
    for (int c0=0;c0<=slices;++c0)      
      glVertex3d(rab*cos(c0*fac),rad*sin(c0*fac),0);      
  glEnd();
}

void glutWireDisc(double rad,int slices) { // ok
  double rab=abs(rad);
  double fac=MATH_2PI/slices;
  glBegin(GL_LINE_STRIP);
    glNormal3d(0,0,math::Sign(rad));
    for (int c0=0;c0<=slices;++c0)      
      glVertex3d(rab*cos(c0*fac),rad*sin(c0*fac),0);      
  glEnd();
}

void glutSolidDiscbw(double rad,int slices,int swp,tensor<double> mas) {
  assert(0<=rad);
  assert(mas.numel==4);
  double fac=MATH_2PI/slices;  
  double lam=0;
  double ang;
  tensor<double> col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, mas(3));
  glBegin(GL_TRIANGLE_FAN);
    glColor4dv(col.data);
    glNormal3d(0,0,math::Sign(rad));
    glVertex3d(0,0,0);
    for (int c0=0;c0<=slices;++c0) {
      if (c0%swp==0)
        lam=1-lam;
      ang=(c0+0.5)*fac;
      glColor4dv((col*lam+mas*(1-lam)).data);
      glVertex3d(rad*cos(ang),rad*sin(ang),0);      
    }
  glEnd();
}

// 3D -----------------------------------

void glutSolidCylinder(double gnd,double hgt,double rad,int slices) {
  double fac=MATH_2PI/slices;
  double cc0,sc0;
  glBegin(GL_QUAD_STRIP);
  for (int c0=0;c0<=slices;++c0) {
    glNormal3d(cc0=cos(c0*fac),sc0=sin(c0*fac),0);
    glVertex3d(rad*cc0,rad*sc0,hgt);
    glVertex3d(rad*cc0,rad*sc0,gnd);
  }
  glEnd();        
}

void glutSolidCylinder(double gnd,double hgt,double rad,double pea,int slices) {
  double fac=MATH_2PI/slices;
  double cc0,sc0;
  glBegin(GL_QUAD_STRIP);
  for (int c0=0;c0<=slices;++c0) {
    glNormal3d(cc0=cos(c0*fac),sc0=sin(c0*fac),0);
    glVertex3d(pea*cc0,pea*sc0,hgt);
    glVertex3d(rad*cc0,rad*sc0,gnd);
  }
  glEnd();        
}

void glutSolidCylinderbw(double gnd,double hgt,double rad,int slices,int swp,tensor<double> mas) {
  assert(mas.numel==4);
  double fac=MATH_2PI/slices;
  double lam=0,cc0,sc0;
  double ang;
  tensor<double> col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, mas(3));
  glBegin(GL_QUAD_STRIP);
    for (int c0=0;c0<=slices;++c0) {
      if (c0%swp==0)
        lam=1-lam;
      ang=(c0+0.5)*fac;
      glNormal3d(cc0=cos(ang),sc0=sin(ang),0);
      glColor4dv((col*lam+mas*(1-lam)).data);
      glVertex3d(rad*cc0,rad*sc0,hgt);
      glVertex3d(rad*cc0,rad*sc0,gnd);
    }
  glEnd();
  glPushMatrix();
    glTranslatef(0,0,hgt);
    glutSolidDiscbw( rad,slices,swp,mas);
  glPopMatrix();
}

void glutSolidSpherebw(double beg,double end,double rad,int slices,int swp,tensor<double> mas) {
  assert(mas.numel==4);
  assert(beg<end);
  assert(0<=rad);
  int cuts=(end-beg)*slices/MATH_2PI;
  double fac=MATH_2PI/slices;
  double mul=(end-beg)/cuts;
  double lam=0;
  tensor<double> col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, mas(3));
  tensor<double> dir(1,3);
  tensor<double> rot=tensor<double>(2,3,3).as(cos(fac),sin(fac),0.0,-sin(fac),cos(fac),0.0,0.0,0.0,1.0);
  for (int c0=0;c0<slices;++c0) {
    if (c0%swp==0)
      lam=1-lam;
    glBegin(GL_QUAD_STRIP);
    for (int c1=0;c1<=cuts;++c1) {
      glColor4dv((col*lam+mas*(1-lam)).data);
      dir.as(cos(c0*fac)*sin(c1*mul+beg),sin(c0*fac)*sin(c1*mul+beg),-cos(c1*mul+beg));
      glNormal3dv(dir.data); glVertex3dv((dir*rad).data);
      dir=rot*dir;
      glNormal3dv(dir.data); glVertex3dv((dir*rad).data);      
    }
    glEnd();
  }
  glPushMatrix();
    glTranslatef(0,0,-rad*cos(end));
    glutSolidDiscbw( rad*sin(end),slices,swp,mas); // TODO increase the radius a bit
  glPopMatrix();
}
