// code by jph

#define HASH_TOTEST   7

//----------------------------------------------------------------------

void tensor_norm2_toTest() {
  tensor<double> vec=tensor<double>::randn(2,3,5);
  printf("%s\n\n",vec.toString());
  printf("%f\n\n",vec.norm2());
  printf("%f\n\n",vec.transpose().norm2());
  tensor<int> in0(1,2); in0.as(0,2);
  tensor<int> in1(1,3); in1.as(1,2,1);
  printf("%s\n\n",vec(in0,in1).toString());
}

void format_txt_toTest() {
  tensor<char> myf=format::txt("ed.txt");
  printf("%sEOF\n",myf.toInfo());
  printf("%sEOF\n",myf.toString());
}

void format_bmp_toTest() {
  tensor<uchar> myf(3,4,10,10);
  myf(0,2,4)=200;
  myf(1,2,4)=200;
  myf(2,2,4)=200;
  myf(3,2,4)=200;
  format::bmp("out/asd.bmp",myf);
}

void blue_toTest(void) {
  time soc=time::now();
  blue<double>   asd(tensor<int>::vector(2).as(100,100),0.025);
  tensor<double> ins(1,2);
  asd.complete(20);
  asd.empty();
  printf("%s\n",asd.toInfo());
  asd.complete(20);
  printf("time == %f \n\n",(time::now()-soc).ms());
  tensor<double> arr=asd.point.toTensor(2);
  arr.store("jans.bin");
  printf("%s\n",asd.toInfo());
}

void hash_flush(const hash<int,float>& myHash,char *name) {
  FILE *fid=fopen(name,"w");
  if (fid) {
    bucket<int,float> buck;
    for (int c0=0;c0<myHash.table.numel;++c0)
      while (myHash.table.data[c0]->token(buck))
        fprintf(fid,"%i\t%i\t%f\n",c0,buck.key,buck.data);
    fclose(fid);
  }
}

void hash_flush(const hash<tensor<int>,tensor<float>>& myHash,char *name) {
  FILE *fid=fopen(name,"w");
  if (fid) {
    bucket<tensor<int>,tensor<float>> buck;
    for (int c0=0;c0<myHash.table.numel;++c0)
      while (myHash.table.data[c0]->token(buck)) {
        printf("%s\n",buck.data.toString());
        fprintf(fid,"[%i]\t\t",c0);
        fprintf(fid,"%s\t\t",buck.key.toString());
        fprintf(fid,"%s\n",buck.data.toString());
      }
    fclose(fid);
  }
}

void tensor_eig_toTest(void) {
  tensor<double> mat=tensor<double>::randn(2,6,6);
  mat=mat+mat.transpose();
  printf("%s\n\n",mat.toString());
  tensor<double> ret;
  //printf("%s\n\n",mat.schur(ret).toString());
  //printf("%s\n\n",ret.toString());
  
  tensor<double> dia=mat.eig(ret);
  printf("%s\n\n",dia.toString());
  printf("%s\n\n",(ret*ret.transpose()).toString());
  printf("%s\n\n",(ret.transpose()*ret).toString());
  printf("%s\n\n",(ret*tensor<double>::diag(dia)*ret.transpose()-mat).toString());
}

void tensor_interp1_toTest(void) {
  tensor<double> asd=tensor<double>::index(1,3);
  tensor<double> itp(1,1);
  asd+=3;
  printf("%s\n\n",asd.toString());  
  for (double c1=-0.4;c1<1.3;c1+=0.03) {
    itp.data[0]=c1;
    printf("%8.3f %f\n",itp.data[0],asd.interp( itp));
  }
}

void tensor_interp2_toTest(void) {
  tensor<double> asd=tensor<double>::index(2,3,2);
  asd+=2;
  tensor<double> itp(1,2);
  printf("%s\n\n",asd.toString());  
  for (double c0=-0.1;c0<=1.1;c0+=0.03) {
    for (double c1=-0.1;c1<=1.1;c1+=0.09) {
      itp.as(c0,c1);
      printf("%5.2f",asd.interp(itp));
    }
    printf("\n");
  } 
  
}

void tensor_interp3_toTest(void) {
  tensor<double> asd=tensor<double>::index(3,3,3,3);
  asd+=3;
  tensor<double> itp(1,3);
  printf("%s\n\n",asd.toString());  
  itp.as(0.25,0.15,0.1);
  itp.as(0.5,0.5,0.51);
  printf("%f\n",asd.interp(itp));
}

void tensor_mean_var_toTest(void) {
  tensor<double> asd=(tensor<double>::rand(2,4,3)*10).to<double>();
  printf("%s\n\n",asd.toString());
  printf("%s\n\n",asd.mean().toString());
  printf("%s\n\n",asd.variance().toString());
}

void tensor_toTest(void) {
  tensor<double> asd=tensor<double>::rand(2,3,3);
  tensor<double> ret;
  tensor<double> qrt=asd.schur(ret);
  printf("%s\n---\n",asd.toString());
  printf("%s\n---\n",ret.chop().toString());
  printf("%s\n---\n",(qrt*ret*qrt.transpose()).toString());
  printf("double\n%s\n", asd.toString());
  printf("int   \n%s\n",(asd*0x7fffffff).to<int>().toString("%08x "));
  printf("--------------------------\n");
  tensor<double> mat=tensor<double>::rand(4,4,3,3,2);
  printf("%s\n\n",mat.toString());
  printf("%s\n\n",mat.cumsize().toString());
  printf("%s\n\n",mat.mean().toString());
  printf("%s\n\n",((mat[0]+mat[1])*0.5).toString());
  assert(mat.mean()==(mat[0]+mat[1])*0.5);
  printf("--------------------------\n");

  //asd.to<double>();
}

template<class type>
void tensor_trace_toTest(void) {
  tensor<int> rnd=(tensor<double>::rand(2,4,4)*4).to<int>();
  printf("%s \n\n",rnd.toString());
  printf("%s \n\n",(rnd==1).toString());
  printf("%s \n\n",(rnd==1).find().toString());
  printf("%s \n",rnd.trace(0,1).toString());
  printf("%i \n",rnd.trace(0,1).data[0]);
}

template<class type>
void tensor_qr_toTest(void) {
  tensor<type> rnd=tensor<type>::randn(2,7,7);
  printf("%s \n\n",rnd.toString());
  tensor<type> ret;
  tensor<type> qor=rnd.qr(ret);
  printf("%s \n\n",qor.toString());
  printf("%s \n\n",ret.toString());
  tensor<type> dif=qor*rnd-ret;
  printf("%s \n\n",dif.toString());
  dif=qor*qor.transpose();
  printf("%s \n\n",dif.toString());
  if (rnd.size[0]==rnd.size[1]) {
    type prd=1;
    for (int c0=0;c0<rnd.size[0];++c0)
      prd*=ret(c0,c0);
    printf("det = %10.2e  %10.2e\n",rnd.det(),prd);
  }
  assert((int)0==0.0);
  assert(qor*qor.transpose()==tensor<type>::eye(qor.size[0])); // qor is orthogonal
  assert(ret==qor*rnd); // R=Q*A <=> Q^t*R=A
  for (int c1=0;c1<rnd.size[1];++c1)
    for (int c0=c1+1;c0<rnd.size[0];++c0)
      assert(abs(ret(c0,c1))<TENSOR_QR); // values below diagonal are null
}

template<class type>
void tensor_svd_sparse_toTest(void) {
  tensor<tensor<type>*> blk(2,2,2);
  //blk(0,0)=new tensor<type>::rand(2,3,3);
  blk(0,0)=new tensor<type>(tensor<type>::rand(2,3,3));
  blk(1,0)=new tensor<type>(2,3,3);
  blk(0,1)=new tensor<type>(2,3,3);
  blk(1,1)=new tensor<type>(2,3,3);
  tensor<type> MAT(blk);
  tensor<type> w,v;  
  tensor<type> d=MAT.svd(w,v);
  printf("%s\n",d.toString());
  printf("%s\n",v.toString());
  printf("%s\n",w.toString());
  printf("------------------\n");
  printf("%s\n",(d.transpose()*tensor<type>::diag(w)*v.transpose()).toString());
  printf("------------------\n");
  printf("%s\n",MAT.toString());
  printf("------------------\n");
}

template<class type>
void tensor_block_toTest(void) {
  tensor<tensor<type>*> bla(2,4,3);
  for (int c0=0;c0<bla.numel;++c0)
    bla.data[c0]=new tensor<type>(2,1,2);
  tensor<type> BLA(bla);
}

template<class type>
void tensor_string_toTest(void) {
  tensor<type> asd=tensor<type>::index(2,2,3)+32;
  printf("%s\n",asd.toString());
  tensor<char> txt=tensor<type>::string((type)123.2);
  printf("-----> %s\n",txt.toString());
}

void tensor_svd_toTest(void) {
  tensor<double> svdt=tensor<double>::randn(2,11,5);
  for (int c1=0;c1<svdt.size[0];c1++)
    svdt(c1,3)=0;
  printf("%s\n\n",svdt.toString());
  tensor<double> uor;
  tensor<double> dia;
  tensor<double> ort;
  tensor<double>::valid=0;
  printf("valid=%i\n",tensor<double>::valid);
  //uor=svdt.svd(dia,ort);
  dia=svdt.svd(uor,ort);
  printf("%s\n\n",dia.toInfo());
  printf("valid=%i\n",tensor<double>::valid);
  printf("%s\n\n",(uor*uor.transpose()).toString());
  printf("%s\n\n",(ort.transpose()*ort).toString());
  printf("%s\n\n",(ort*ort.transpose()).toString());
  dia=tensor<double>::diag(dia);
  printf("%s\n\n",dia.toString());
  printf("difference=\n%s\n\n",(uor.transpose()*dia*ort.transpose()-svdt).toString());
  //printf("%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
}

template<class type>
void tensor_lu_toTest() {
  tensor<int> ind;
  type sgn;
  tensor<type> mlu=tensor<type>::randn(2,4,4);
  tensor<type> rhs=tensor<type>::randn(1,4);
  tensor<type> ret=mlu.lu(ind,sgn);
  tensor<type> sol=ret.lu(ind,rhs);
  
  printf("%s\n\n",(mlu*sol).toString());
  printf("%s\n\n",rhs.toString());
  
  printf("%s\n\n",sol.toString());
  printf("%s\n\n",mlu.solve(rhs).toString());
}

template<class type>
void tensor_inverse_toTest() {
  tensor<type> rnd=tensor<type>::rand(2,7,5);
  tensor<type> spl=tensor<type>::eye(5);
  spl(4,3)=spl(3,4)=spl(1,2)=spl(2,1)=1;
  rnd=rnd*spl;
  //rnd=rnd.transpose();
  printf("rank = %i\n\n",rnd.rank());
  printf("FOR MATLAB\n%s\n---------\n",rnd.toString());
  tensor<type> ide=rnd.inverse();
  printf("OUR PINVERSE:\n%s\n--------------\n",ide.toString());
  //printf("%s\n",ide.toString());
  //tensor<type> fgh=tensor<type>::randn(2,7,7);
  //tensor<type> dif=fgh.solve(tensor<type>::eye(7))-fgh.inverse();
  //printf("inv=pinv :\n%s\n",dif.toString());
}

void tensor_rank_toTest(void) {
  tensor<double> svdt=tensor<double>::randn(2,11,11);
  for (int c1=0;c1<svdt.size[0];c1++) {    
    //svdt(c1,1)=TENSOR_NO_DIVISION;
    //svdt(c1,3)=0;
  }
  printf("rank = %i\n\n",svdt.transpose().rank());
}

void tensor_sort_inv_test() {
  tensor<int> rnd=(tensor<float>::rand(1,15)*100).to<int>();
  tensor<int> ind;
  tensor<int> srt(ind);
  //tensor<int> srt=rnd.sort(ind);
  printf("%s\n",srt.toString());
}

void tensor_svd_toTest2(void) {
  //tensor<double> svdt=tensor<double>::randn(2,6,6);
  tensor<double> svdt(2,5,5);//=tensor<double>::randn(2,6,6);
  //for (int c1=0;c1<svdt.size[0];c1++) {
  //  svdt(c1,3)=0;
  //  svdt(c1,4)=0;
  //  svdt(c1,5)=0;
  //}
  printf("%s\n\n",svdt.toString());
  tensor<double> uor;
  tensor<double> dia;
  tensor<double> ort;
  tensor<double>::valid=0;
  printf("valid=%i\n",tensor<double>::valid);
  uor=svdt.svd(dia,ort);
  printf("valid=%i\n",tensor<double>::valid);
  printf("%s\n\n",uor.toString());
  printf("%s\n\n",(uor.transpose()*uor).toString());
  printf("%s\n\n",(ort.transpose()*ort).toString());
  printf("%s\n\n",(ort*ort.transpose()).toString());
  dia=tensor<double>::diag(dia);
  printf("%s\n\n",dia.toString());
  printf("difference=\n%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
  //printf("%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
}

void sha_toTest() {
  sha mySha;
  char* stra=new char[16];
  stra[ 0]=0x0;
  stra[ 1]=0x1;
  stra[ 2]=0x2;
  stra[ 3]=0x3;
  stra[ 4]=0x4;
  stra[ 5]=0x5;
  stra[ 6]=0x6;
  stra[ 7]=0x7;
  stra[ 8]=0x8;
  stra[ 9]=0x9;
  stra[10]=0x10;
  stra[11]=0x11;
  stra[12]=0x12;
  stra[13]=0x13;
  stra[14]=0x14;
  stra[15]=0x15;
  mySha.add(stra,16);
  mySha.add(stra,16);
  mySha.add(stra,16);
  mySha.add(stra,16);
  //mySha.add(stra,16);
  //mySha.add(stra,10);
  //mySha.add(stra,10);
  //mySha.add(stra,10);
  printf("%s\n",mySha.toInfo());
  printf("%s\n",mySha.toString());
  //tensor<uint> asd2=mySha.flush();
  tensor<int> asd=mySha.flush();
  printf("\n%s\n3fc78b03 7eb94ea8 d0cc96b2 c1a172b7 3df20a50\n",asd.toString("%08x "));
  //printf("%s\n",mySha.toInfo());
  //printf("%s\n",mySha.toString());
  //printf("%08x\n",asd.sum());
  //mySha.add(stra,3);
  //printf("%s\n",mySha.toInfo());
  //asd=mySha.flush();  
  //printf("%s\n",asd.toString("%08x "));
  //printf("%08x\n",asd.sum());
  //tensor<int> dist(1,13);  
  //for (int c1=0;c1<10000;c1++) {
  //  tensor<double> some=tensor<double>::rand(2,10,4);
  //  //printf("%s\n",some.toInfo());
  //  //printf("%s\n",some.toString());
  //  //printf("%08x\n",sha_hash(some));
  //  //printf("%08x\n",sha::hash<double>(some));
  //  //printf("%i\n",sha::hash<double>(some));
  //  dist.data[sha::hash<double>(some)%13]++;
  //}
  //for (int c1=0;c1<13;c1++) 
  //  printf("%3i %i\n",c1,dist.data[c1]);
}

void random_toTest(void) {
  //random<double>::randomize();
  printf("%s\n",random<double>::toInfo());
  double d1=random<double>::uniform();
  printf("%s\n",random<double>::toInfo());
  double n1,n2;
  for (int c1=0;c1<10;c1++) {
    n1=random<double>::normal();
    printf("%f\n",n1);
  }
}

//----------------------------------------------------------------------

void complex_toTest() {
  complex<double> a(-2,3);
  complex<double> b(5,9);
  printf("%s\n",(a/b).toInfo());
  printf("%s\n",(a*b).toInfo());
}

void node_toTest(void) {
  node<int> myNodeA(0,3);
  node<int> myNodeB(&myNodeA,2);
  printf("%s\n",myNodeA.toInfo());
  printf("%s\n",myNodeB.toInfo());
}

//----------------------------------------------------------------------

void linked_toTest(void) {
  int myInt;
  linked<int> myLinked;
  printf("%s\n",myLinked.toInfo());
  myLinked.add(0);
  printf("%s <- 1 added\n",myLinked.toInfo());
  while (myLinked.token(myInt))
   myLinked.remove();
  printf("%s <- 1 removed \n",myLinked.toInfo());
  myLinked.add(1);
  myLinked.add(2);
  myLinked.add(3);
  myLinked.add(4);
  myLinked.add(5);
  myLinked.add(6);
  printf("%s\n",myLinked.toInfo());
  
  while (myLinked.token(myInt)) {
    printf("token %i\n",myInt);
    if (myInt==3||myInt==1) {
      myLinked.remove();
      myLinked.add(0);      
      myLinked.pre(-1);
    }
  }
  printf("%s\n",myLinked.toInfo());
  while (myLinked.token(myInt))
    printf("token %i\n",myInt);
  //myLinked.token_reset();
  printf("----\n");
  printf("%s\n",myLinked.toInfo());  
  while (myLinked.token(myInt)) {
    printf("token %i\n",myInt);
  }
  tensor<int> myTensor=myLinked.toTensor();
  printf("tensor %s\n",myTensor.toString());
  printf("%s\n",myLinked.toInfo());
  while (myLinked.token(myInt))
    myLinked.remove();    
  printf("%s <- after while remove \n",myLinked.toInfo());  
  myLinked.empty();
  printf("%s\n",myLinked.toInfo());  
  printf("-------\n");
  linked<int>* myStar=new linked<int>();
  printf("%s\n",myStar->toInfo());
  myStar->add(0);
  myStar->add(1);
  printf("%s\n",myStar->toInfo());
  delete myStar;
  //printf("%s\n",myStar->toInfo());  
}

int hash_toTest_int(int myInt) {
  return myInt%HASH_TOTEST;
}

void hash_toTest_int(void) {
  hash<int,float> myHash(HASH_TOTEST,set<int>::map);  
  float myFloat=0.23f;
  myHash.add(3,myFloat);
  myFloat=0;
  myHash.add(3,myFloat);
  printf("float %8.3f\n",myFloat);
  myFloat=-10;
  myHash.add(3+5,myFloat);
  myFloat=.3333;
  myHash.add(3+25,myFloat);
  bucket<int,float> buck;
  for (int c0=0;c0<myHash.table.numel;++c0)
    while (myHash.table(c0)->token(buck))
      printf("%i key=%4i data=%8.3f\n",c0,buck.key,buck.data);
  printf("%s\n",myHash.toInfo());
  int status;
  status=myHash.get(3+2,myFloat);
  printf("get 5 %i %6.3f\n",status,myFloat);
  status=myHash.get(3+5,myFloat);
  printf("get 8 %i %6.3f\n",status,myFloat);
  hash_flush(myHash,"myhash.txt");  
}

void hash_toTest_void(void) {
  hash<int,void*> myHash(HASH_TOTEST,set<int>::map);  
  void* null=(void*)9999;
  myHash.add(3,null);
  int status;
  status=myHash.get(3,null);
  printf("get 3 %i %08x\n",status,null);
}


int hash_toTest_tensor(tensor<int> myTensor) {
  return myTensor.sum()%HASH_TOTEST;
}

void hash_toTest_tensor(void) {
  hash<tensor<int>,tensor<float>> myHash(HASH_TOTEST,set<tensor<int>>::map);
  tensor<int> key(1,3);
  key.as(2,3,4);
  tensor<float> color(1,3);
  color(0)=0.222;
  color(1)=0.111;
  color(2)=0.333;
  myHash.add(key,color);
  key(1)++;
  myHash.add(key,color);
  
  printf("%s\n",myHash.toInfo());
  hash_flush(myHash,"myhash_tensor.txt");
}

void set_toTest(void) {
  set<int> myS(10,set<int>::map);
  myS.toggle(3);
  myS.toggle(10);
  myS.toggle(13);
  printf("%s\n",myS.toInfo());
  myS.toggle(3);
  myS.toggle(10);
  myS.toggle(13);  
  printf("%s\n",myS.toInfo());
  
  set<tensor<int>> myV(11,set<tensor<int>>::map);
  tensor<int> inp(1,3);
  inp.data[2]=1;
  myV.toggle(inp);
  printf("%s\n",myV.toInfo());
  inp.data[1]=2;
  myV.toggle(inp);
  printf("%s\n",myV.toInfo());  
}

void hash_toTest_newest(void) {
  hash<int,int> myV(1,set<int>::map);
  myV.add(123,3);
  myV.add(12,144);
  printf("%s\n",myV.toString());
}

void tensor_toTest_char(void) {
  tensor<char> st1=tensor<char>::string("get/meshgl/jaja");
  tensor<char> st2=tensor<char>::string("get/meshgl/jaj");
  printf("%i\n",st1==st2);
  printf("%s\n",st1.toString());
  
}
