// code by jph
// after Bridson blue noise, Siggraph 2007

#ifndef RANDOM_BLUE_INCLUDED
#define RANDOM_BLUE_INCLUDED

static char BLUE_BUFFER[8192];

template<class type>
class blue {
public:
  blue<type>(tensor<int> size,type radius); // only constructor
  
  int insert(tensor<type> pnt);
  void complete(int num);
  void empty(void);
  tensor<type> extension(void);
  
  char* toString(void);
  char* toInfo(void);

  tensor<int> grid; // value 0 = unoccupied, -1 = blocked
  hash<uint,tensor<type>> point; // otherwise corresponding point is stored in tab
  type rad;
  type len;
  
  tensor<int> blk; // like statics TODO
  tensor<int> chk; // like statics
};

template<class type>
blue<type>::blue(tensor<int> size,type radius) :
  grid(size.numel,size.data) ,
  point(size.prod()/(1<<(size.numel+2))) ,
  rad(radius) ,
  len(radius/sqrt((type)size.numel)) {
  assert(size.numel==2); // mmh... a little dangerous!!!
  // TODO these are statics
  switch (size.numel) {
    case 2:
      blk=tensor<int>(2,2,8).as(
        -1,-1,
        -1, 0,
        -1, 1,
         0,-1,
         0, 1,
        +1,-1,
        +1, 0,
        +1, 1);
      chk=tensor<int>(2,2,36).as( 
         2, 0,
         3, 0,
         2, 1,
         3, 1,
         1, 2,
         2, 2,
         3, 2,
         1, 3,
         2, 3,
        -0, 2, // ---
        -0, 3,
        -1, 2,
        -1, 3,
        -2, 1,
        -2, 2,
        -2, 3,
        -3, 1,
        -3, 2,          
        -2,-0, // ---
        -3,-0,
        -2,-1,
        -3,-1,
        -1,-2,
        -2,-2,
        -3,-2,
        -1,-3,
        -2,-3,
         0,-2, // ---
         0,-3,
         1,-2,
         1,-3,
         2,-1,
         2,-2,
         2,-3,
         3,-1,
         3,-2);
      break;
  }
}

template<class type>
int blue<type>::insert(tensor<type> pnt) {
  assert(grid.ndims<=pnt.numel);
  int c0;
  int flg,tst,ind;
  tensor<int> cum=grid.cumsize();
  tensor<type> ref;
  flg=1;
  for (c0=0;c0<pnt.numel;++c0)
    flg&=0<=pnt.data[c0];
  if (flg) {
    tensor<int> idx=(pnt/len).to<int>();
    ind=grid.at(idx,cum);
    if (0<=ind)
      if (grid.data[ind]==0) { // grid cell of point unoccupied
        flg=1;
        for (c0=0;c0<blk.size[1];++c0) { // check blocking range
          tst=grid.at(idx+blk[c0],cum);
          if (0<=tst)
            flg&=grid.data[tst]<=0;
        }
        if (flg) {
          for (c0=0;c0<chk.size[1];++c0) {
            tst=grid.at(idx+chk[c0],cum);
            if (0<=tst)
              if (0<grid.data[tst]) {
                point.get(grid.data[tst],ref);
                flg&=2*rad<(pnt-ref).norm();
              }
          }
          if (flg) {
            point.add(point.numel+1,pnt);
            return grid.data[ind]=point.numel;
          }
        }
      }    
  }
  return 0;
}

template<class type>
void blue<type>::complete(int num) {
  assert(grid.ndims==2);
  int ind=0;
  type ang,exc;
  tensor<type> pnt(1,grid.ndims);
  tensor<type> ref;
  if (!point.numel) { // TODO
    for (int c0=0;c0<grid.ndims;++c0)
      pnt.data[c0]=(random<double>::uniform()*0.5+0.25)*grid.size[c0]*len;
    insert(pnt);
  }
  while (point.get(++ind,ref)) {
    ang=random<type>::uniform()*MATH_2PI;
    for (int c0=0;c0<num;++c0) {
      ang+=1;
      exc=(1+random<type>::uniform())*2*rad; // TODO this is not according to the measure
      pnt.data[0]=ref.data[0]+cos(ang)*exc;
      pnt.data[1]=ref.data[1]+sin(ang)*exc;
      insert(pnt);
    }
  }
}

template<class type>
void blue<type>::empty(void) {
  for (int c0=0;c0<grid.numel;++c0)
    if (0<grid.data[c0])
      grid.data[c0]=0;
  point.empty();
}

template<class type>
tensor<type> blue<type>::extension(void) {
  tensor<type> ret(1,grid.ndims);
  for (int c0=0;c0<grid.ndims;++c0)
    ret.data[c0]=grid.size[c0]*len;
  return ret;
}

template<class type>
char* blue<type>::toString(void) {
  tensor<int>::format="%i";
  char* ret=(!!grid).toString();
  tensor<int>::format="%i ";
  return ret;
}

template<class type>
char* blue<type>::toInfo(void) {
  sprintf(BLUE_BUFFER,"blue<%i>([%10.2e,%10.2e] rad=%10.2e len=%10.2e numel=%i/%i)",sizeof(type),grid.size[0]*len,grid.size[1]*len,rad,len,point.numel,point.table.numel);
  return BLUE_BUFFER;
}

  
#endif
