// code by jph

// since there is no copyconstructor, or operator= defined
//  => always use  const rapid&  when passing!

#include "../_extern/RAPID.H"

class rapid {
public:
  rapid(void);
  ~rapid(void);
  void set(meshgl& msh);
  
  RAPID_model* model;
  meshgl mesh; // TODO make this a pointer, manage as in opengl!!
  tensor<double> area;
  tensor<double> V3D; // 3xT avg points of shape
  tensor<double> N3D; // 3xT avg normal or shape

  static int collide(tensor<double>& pos0,const rapid& rap0,tensor<double>& pos1,const rapid& rap1,int flag);
  static tensor<int>    pair;
  static tensor<int>    hits[2];
  static tensor<double> pnt;
  static tensor<double> nrm;
  static double         str;
  static int            num_contacts;
};

rapid::rapid(void) { 
  model=0;
}
  
inline rapid::~rapid(void) {
  if (model)
    delete model;
}

void rapid::set(meshgl& msh) {
  if (model)
    delete model;
  model=new RAPID_model;
  model->BeginModel();
  mesh=msh;
  area=tensor<double>(1,mesh.ind[MESH_TRIA].size[1]);
  V3D=tensor<double>(2,3,mesh.ind[MESH_TRIA].size[1]);
  N3D=tensor<double>(2,3,mesh.ind[MESH_TRIA].size[1]);
  int ofs=mesh.offsetV3F();
  tensor<double> v0(1,3),v1(1,3),v2(1,3),crs;
  for (int c0=0;c0<mesh.ind[MESH_TRIA].size[1];++c0) {
    v0.in(0,&mesh.arr[mesh.ind[MESH_TRIA](0,c0)].to<double>().data[ofs],3);
    v1.in(0,&mesh.arr[mesh.ind[MESH_TRIA](1,c0)].to<double>().data[ofs],3);
    v2.in(0,&mesh.arr[mesh.ind[MESH_TRIA](2,c0)].to<double>().data[ofs],3);
    model->AddTri(v0.data,v1.data,v2.data,c0);
    crs=(v1-v0).x(v2-v0);
    area(c0)=crs.norm()*0.5;
    V3D.in(c0*3,((v0+v1+v2)*MATH_1_3).data,3);
    N3D.in(c0*3,        crs.normal().data,3);
  }
  model->EndModel();
}

int rapid::collide(tensor<double>& pos0,const rapid& rap0,
                    tensor<double>& pos1,const rapid& rap1,int flag) {
  assert(rap0.model&&rap1.model);
  assert(rap0.model!=rap1.model);
  int c0,id0,id1;
  double rot0[3][3];
  double rot1[3][3];
  rot0[0][0]=pos0.data[ 0]; rot0[0][1]=pos0.data[ 4]; rot0[0][2]=pos0.data[ 8];
  rot0[1][0]=pos0.data[ 1]; rot0[1][1]=pos0.data[ 5]; rot0[1][2]=pos0.data[ 9];
  rot0[2][0]=pos0.data[ 2]; rot0[2][1]=pos0.data[ 6]; rot0[2][2]=pos0.data[10];
  rot1[0][0]=pos1.data[ 0]; rot1[0][1]=pos1.data[ 4]; rot1[0][2]=pos1.data[ 8];
  rot1[1][0]=pos1.data[ 1]; rot1[1][1]=pos1.data[ 5]; rot1[1][2]=pos1.data[ 9];
  rot1[2][0]=pos1.data[ 2]; rot1[2][1]=pos1.data[ 6]; rot1[2][2]=pos1.data[10];
  RAPID_Collide(rot0,pos0[3].data,rap0.model,
                rot1,pos1[3].data,rap1.model,flag);
  rapid::num_contacts=RAPID_num_contacts;
  pair=tensor<int>(2,num_contacts,2);
  for (c0=0;c0<num_contacts;++c0) {
    pair(c0,0)=RAPID_contact[c0].id1;
    pair(c0,1)=RAPID_contact[c0].id2;
  }
  hits[0]=pair[0].unique();
  hits[1]=pair[1].unique();
  tensor<double>::valid=1;
  pnt=0;
  nrm=0;
  str=0;
  if (num_contacts) {
    double sca;
    tensor<double> nrm0,nrm1;
    tensor<double> aff(1,4); aff(3)=1;
    tensor<double> vec0(1,4);
    tensor<double> vec1(1,4);
    tensor<double> pnt_(1,4);
    tensor<double> nrm_(1,4);
    for (c0=0;c0<num_contacts;++c0) {
      id0=pair(c0,0);
      id1=pair(c0,1);
      if (rap0.area.data[id0]<rap1.area.data[id1]) {
        assert(id0*3<rap0.V3D.numel);
        pnt_+=pos0*aff.in(0,&rap0.V3D.data[id0*3],3);
      } else {      
        assert(id1*3<rap1.V3D.numel);
        pnt_+=pos1*aff.in(0,&rap1.V3D.data[id1*3],3);
      }
      nrm0=pos0*vec0.in(0,&rap0.N3D.data[id0*3],3);
      nrm1=pos1*vec1.in(0,&rap1.N3D.data[id1*3],3);
      sca=-min(0,nrm0|nrm1)*min(rap0.area.data[id0],
                                rap1.area.data[id1]);
      nrm_+=(nrm1-nrm0)*sca;
      str+=sca;
    }
    rapid::pnt=pnt_*(1/pnt_(3));
    rapid::nrm=nrm_.normal();
  }
  return num_contacts&&tensor<double>::valid;
}

tensor<int>     rapid::pair=tensor<int>(2,0,2);
tensor<int>     rapid::hits[2]={tensor<int>(1,0),tensor<int>(1,0)};
tensor<double>  rapid::pnt=tensor<double>(1,4);
tensor<double>  rapid::nrm=tensor<double>(1,4);
double          rapid::str=0;                         // TODO later merge with nrm
int             rapid::num_contacts=0;

